const builtin = @import("builtin");

const coral = @import("./coral.zig");

const std = @import("std");

pub fn CallTask(comptime function: anytype) type {
    const Function = @TypeOf(function);

    const function_fn = switch (@typeInfo(Function)) {
        .@"fn" => |@"fn"| @"fn",
        else => @compileError("expeceted param `exec` to be an fn type"),
    };

    const Args = std.meta.ArgsTuple(Function);

    return struct {
        args: Args,
        @"return": Future(function_fn.return_type.?),

        const Self = @This();

        pub fn init(args: Args) Self {
            return .{
                .args = args,
                .@"return" = .empty,
            };
        }

        pub fn run(self: *Self) void {
            std.debug.assert(self.@"return".resolve(@call(.auto, function, self.args)));
        }
    };
}

pub fn BlockingQueue(comptime max: usize, comptime T: type) type {
    const circular_max = max + 1;

    return struct {
        buffer: [circular_max]T,
        head: usize,
        tail: usize,
        access: std.Thread.Mutex,
        not_empty: std.Thread.Condition,
        not_full: std.Thread.Condition,

        const Self = @This();

        pub fn dequeue(self: *Self) T {
            self.access.lock();

            defer {
                self.access.unlock();
            }

            while (self.head == self.tail) {
                self.not_empty.wait(&self.access);
            }

            const item = self.buffer[self.head];
            const wrapped_head = (self.head + 1) % circular_max;

            self.head = wrapped_head;

            self.not_full.signal();

            return item;
        }

        pub const empty = Self{
            .buffer = undefined,
            .head = 0,
            .tail = 0,
            .access = .{},
            .not_empty = .{},
            .not_full = .{},
        };

        pub fn enqueue(self: *Self, item: T) void {
            self.access.lock();

            defer {
                self.access.unlock();
            }

            const wrapped_tail = (self.tail + 1) % circular_max;

            while (wrapped_tail == self.head) {
                self.not_full.wait(&self.access);
            }

            self.buffer[self.tail] = item;
            self.tail = wrapped_tail;

            self.not_empty.signal();
        }
    };
}

pub const Event = Future(void);

pub fn Future(comptime Payload: type) type {
    return struct {
        payload: ?Payload,
        resolved: std.Thread.ResetEvent,

        const Self = @This();

        pub const empty = Self{
            .payload = null,
            .resolved = .{},
        };

        pub fn get(self: *Self) *Payload {
            // TODO: Make async.
            self.resolved.wait();

            return &self.payload.?;
        }

        pub fn resolve(self: *Self, payload: Payload) bool {
            if (self.resolved.isSet()) {
                return false;
            }

            self.payload = payload;

            self.resolved.set();

            return true;
        }
    };
}

pub const TaskQueue = struct {
    threads: []std.Thread,
    arena: std.heap.ArenaAllocator,
    tasks: *Buffer,

    const Buffer = struct {
        waiting: std.Thread.WaitGroup,
        pending: BlockingQueue(512, Runnable),

        fn consume(self: *Buffer) void {
            while (true) {
                const runner = self.pending.dequeue();

                if (runner.is_fn(poison_pill)) {
                    break;
                }

                runner.call(.{});
                self.waiting.finish();
            }
        }

        fn poison(self: *Buffer, threads: []const std.Thread) void {
            var threads_remaining = threads.len;

            while (threads_remaining != 0) : (threads_remaining -= 1) {
                self.pending.enqueue(.initFn(poison_pill));
            }

            for (threads) |thread| {
                thread.join();
            }
        }

        fn poison_pill() void {}
    };

    const Runnable = coral.Callable(void, &.{});

    pub fn create(self: *TaskQueue, task: anytype) std.mem.Allocator.Error!*const @TypeOf(task) {
        const Task = @TypeOf(task);

        switch (@hasDecl(Task, "run")) {
            true => switch (@TypeOf(Task.run)) {
                fn (*Task) void, fn (*const Task) void => {},
                else => @compileError("param `task` must contain a `run` declaration"),
            },

            else => @compileError("`run` declaration on param `task` must be an fn type"),
        }

        switch (builtin.single_threaded) {
            true => {
                Runnable.initRef(&task, task.run).call(.{});

                return &task;
            },

            else => {
                const allocated_task = try self.arena.allocator().create(Task);

                allocated_task.* = task;

                self.tasks.waiting.start();
                self.tasks.pending.enqueue(.initRef(allocated_task, Task.run));

                return allocated_task;
            },
        }
    }

    pub fn execute(self: *TaskQueue, task: anytype) std.mem.Allocator.Error!void {
        _ = try self.create(task);
    }

    pub fn deinit(self: *TaskQueue) void {
        self.tasks.waiting.wait();
        self.tasks.poison(self.threads);
        self.arena.deinit();
        coral.heap.allocator.destroy(self.tasks);
        coral.heap.allocator.free(self.threads);

        self.* = undefined;
    }

    pub fn finish(self: *TaskQueue) void {
        self.tasks.waiting.wait();
        self.tasks.waiting.reset();

        _ = self.arena.reset(.retain_capacity);
    }

    pub fn init(thread_count: usize) (std.mem.Allocator.Error || std.Thread.SpawnError)!TaskQueue {
        const tasks = try coral.heap.allocator.create(Buffer);

        errdefer {
            coral.heap.allocator.destroy(tasks);
        }

        tasks.* = .{
            .pending = .empty,
            .waiting = .{},
        };

        const threads = try coral.heap.allocator.alloc(std.Thread, if (builtin.single_threaded) 0 else thread_count);
        var threads_spawned: usize = 0;

        errdefer {
            tasks.poison(threads[0..threads_spawned]);
            coral.heap.allocator.free(threads);
        }

        for (threads) |*thread| {
            thread.* = try .spawn(.{}, Buffer.consume, .{tasks});
            threads_spawned += 1;
        }

        std.debug.assert(threads_spawned == thread_count);

        const name = comptime try coral.ShortString.init("ona worker");

        for (threads) |thread| {
            thread.setName(name.slice()) catch |set_name_error| {
                switch (set_name_error) {
                    error.Unsupported, error.NameTooLong => break,
                    else => continue,
                }
            };
        }

        return .{
            .arena = .init(coral.heap.allocator),
            .tasks = tasks,
            .threads = threads,
        };
    }
};

pub fn wait_all(comptime Payload: type, futures: []const *Future(Payload)) void {
    for (futures) |future| {
        _ = future.get();
    }
}
