pub const Field = @import("./shaders/Field.zig");

pub const Root = @import("./shaders/Root.zig");

pub const Scope = @import("./shaders/Scope.zig");

pub const Type = @import("./shaders/Type.zig");

const coral = @import("./coral.zig");

const glsl = @import("./shaders/glsl.zig");

const spirv = @import("./shaders/spirv.zig");

const std = @import("std");

const tokens = @import("./shaders/tokens.zig");

pub const Argument = struct {
    expression: *const Expression,
    has_next: ?*const Argument = null,
};

pub const Block = struct {
    scope: *const Scope,
    depth: usize,
    has_statement: ?*const Statement = null,
};

pub const DefinitionError = std.mem.Allocator.Error || error{
    TooManySymbols,
    DuplicateIdentifier,
};

pub const Expression = union(enum) {
    float: Float,
    int: Int,
    group_expression: *const Expression,
    add: BinaryOperation,
    subtract: BinaryOperation,
    multiply: BinaryOperation,
    divide: BinaryOperation,
    equal: BinaryOperation,
    greater_than: BinaryOperation,
    greater_equal: BinaryOperation,
    lesser_than: BinaryOperation,
    lesser_equal: BinaryOperation,
    negate_expression: *const Expression,
    get_local: *const Local,
    get_parameter: *const Parameter,
    get_uniform: GetUniform,
    get_object: GetObject,
    mutate_local: LocalMutation,
    mutate_output: OutputMutation,
    get_texture: *const Texture,
    get_output: *const Output,
    get_input: *const Input,
    invoke: Invocation,
    convert: Conversion,
    pow: Intrinsic,
    abs: Intrinsic,
    sin: Intrinsic,
    sample: Builtin,

    pub const BinaryOperation = struct {
        rhs_expression: *const Expression,
        lhs_expression: *const Expression,

        pub fn inferType(self: BinaryOperation) TypeError!*const Type {
            const lhs_type = try self.lhs_expression.inferType();
            const rhs_type = try self.rhs_expression.inferType();

            if (lhs_type != rhs_type) {
                return error.IncompatibleTypes;
            }

            return lhs_type;
        }
    };

    pub const Builtin = struct {
        has_argument: ?*const Argument,
        signatures: []const Signature,

        pub const Signature = struct {
            parameter_types: []const *const Type,
            return_type: *const Type,
        };

        pub fn inferType(self: Builtin) TypeError!*const Type {
            signature_matching: for (self.signatures) |signature| {
                var parameter_index: usize = 0;
                var has_argument = self.has_argument;

                while (has_argument) |argument| : ({
                    has_argument = argument.has_next;
                    parameter_index += 1;
                }) {
                    if (signature.parameter_types.len == parameter_index) {
                        continue :signature_matching;
                    }

                    if (signature.parameter_types[parameter_index] != try argument.expression.inferType()) {
                        continue :signature_matching;
                    }
                }

                return signature.return_type;
            }

            return error.IncompatibleArguments;
        }
    };

    pub const Conversion = struct {
        target_type: *const Type,
        first_argument: *const Argument,
        parameter_types: []const *const Type,
    };

    pub const Float = struct {
        whole: [:0]const u8,
        decimal: [:0]const u8,
    };

    pub const Intrinsic = struct {
        allowed_parameter_types: []const *const Type,
        expected_parameter_count: usize,
        first_argument: *const Argument,

        pub fn inferType(self: Intrinsic) TypeError!*const Type {
            const return_type = try self.first_argument.expression.inferType();

            if (std.mem.indexOfScalar(*const Type, self.allowed_parameter_types, return_type) == null) {
                return error.IncompatibleTypes;
            }

            var has_next_argument = self.first_argument.has_next;
            var arguments_remaining = self.expected_parameter_count;

            while (has_next_argument) |next_argument| : ({
                has_next_argument = next_argument.has_next;
                arguments_remaining += 1;
            }) {
                if (arguments_remaining == 0) {
                    return error.IncompatibleArguments;
                }

                if (try next_argument.expression.inferType() != return_type) {
                    return error.IncompatibleTypes;
                }
            }

            if (arguments_remaining != 0) {
                return error.IncompatibleArguments;
            }

            return return_type;
        }
    };

    pub const GetObject = struct {
        field: *const Field,
        object_expression: *const Expression,
    };

    pub const GetUniform = struct {
        field: *const Field,
        uniform: *const Uniform,
    };

    pub const Invocation = struct {
        function: *const Function,
        has_argument: ?*const Argument = null,

        pub fn inferType(self: Invocation) TypeError!*const Type {
            var has_parameter = self.function.has_parameter;
            var has_argument = self.has_argument;

            while (has_parameter) |parameter| {
                const argument = has_argument orelse {
                    return error.IncompatibleArguments;
                };

                defer {
                    has_parameter = parameter.has_next;
                    has_argument = argument.has_next;
                }

                if (parameter.type != try argument.expression.inferType()) {
                    return error.IncompatibleTypes;
                }
            }

            if (has_argument != null) {
                return error.IncompatibleArguments;
            }

            return self.function.has_return_type orelse error.IncompatibleTypes;
        }
    };

    pub const Int = struct {
        literal: [:0]const u8,
    };

    pub const OutputMutation = struct {
        output: *const Output,
        expression: *const Expression,
    };

    pub const LocalMutation = struct {
        local: *const Local,
        expression: *const Expression,
    };

    pub fn inferType(self: Expression) TypeError!*const Type {
        return switch (self) {
            .float => .float,
            .int => .int,
            .negate_expression, .group_expression => |expression| expression.inferType(),
            .abs, .pow, .sin => |generic| generic.inferType(),
            .convert => |convert| convert.target_type,
            .invoke => |invoke| invoke.inferType(),
            .mutate_local => |mutate_local| mutate_local.local.type,
            .get_local => |local| local.type,
            .get_object => |get_object| get_object.field.type,
            .get_parameter => |get_parameter| get_parameter.type,
            .sample => |builtin| builtin.inferType(),
            .get_uniform => |uniform| uniform.field.type,
            .get_input => |input| input.type,
            .get_output => |output| output.type,
            .mutate_output => |mutate_output| mutate_output.output.type,

            .get_texture => |texture| switch (texture.layout) {
                .dimensions_2 => .texture2,
            },

            .add,
            .subtract,
            .multiply,
            .divide,
            .equal,
            .greater_than,
            .greater_equal,
            .lesser_than,
            .lesser_equal,
            => |binary_op| binary_op.inferType(),
        };
    }
};

pub const Function = struct {
    identifier: [:0]const u8,
    has_return_type: ?*const Type = null,
    has_parameter: ?*const Parameter = null,
    block: *const Block,
};

pub const GenerationError = coral.bytes.ReadWriteError || error{
    UnsupportedFeature,
};

pub const Input = struct {
    identifier: [:0]const u8,
    type: *const Type,
    location: u8,
};

pub const Output = struct {
    identifier: [:0]const u8,
    type: *const Type,
    location: u8,
};

pub const Parameter = struct {
    identifier: [:0]const u8,
    type: *const Type,
    has_next: ?*const Parameter = null,
};

pub const ParsingError = std.mem.Allocator.Error || tokens.ExpectationError || DefinitionError || TypeError || error{
    ImmutableStorage,
    UndefinedIdentifier,
};

pub const Texture = struct {
    identifier: [:0]const u8,
    binding: u8,
    layout: Layout,

    pub const Layout = enum {
        dimensions_2,
    };
};

pub const Statement = union(enum) {
    declare_local: LocalDeclaration,
    mutate_local: LocalMutation,
    mutate_output: OutputMutation,
    return_expression: ?*const Expression,

    pub const LocalDeclaration = struct {
        local: *const Local,
        has_next: ?*const Statement = null,
    };

    pub const LocalMutation = struct {
        local: *const Local,
        expression: *const Expression,
        has_next: ?*const Statement = null,
    };

    pub const OutputMutation = struct {
        output: *const Output,
        expression: *const Expression,
        has_next: ?*const Statement = null,
    };
};

pub const Local = struct {
    identifier: [:0]const u8,
    expression: *const Expression,
    type: *const Type,
    is_constant: bool,
};

pub const TypeError = error{
    IncompatibleTypes,
    IncompatibleArguments,
};

pub const Uniform = struct {
    identifier: [:0]const u8,
    semantic: [:0]const u8,
    binding: u8,
    has_field: ?*const Field = null,
};

pub const generateGlsl = glsl.generate;

pub const generateSpirv = spirv.generate;
