const coral = @import("../coral.zig");

const std = @import("std");

const tokens = @import("./tokens.zig");

scope: *coral.shaders.Scope,
inputs: std.BoundedArray(*const coral.shaders.Input, 15) = .{},
outputs: std.BoundedArray(*const coral.shaders.Output, 15) = .{},
uniforms: std.BoundedArray(*const coral.shaders.Uniform, 15) = .{},
textures: std.BoundedArray(*const coral.shaders.Texture, 15) = .{},
functions: std.BoundedArray(*const coral.shaders.Function, 255) = .{},

pub const ParseResult = union(enum) {
    ok,
    failure: [:0]const u8,
};

const Self = @This();

pub fn clear(self: *Self) void {
    self.scope.clear();
    self.inputs.clear();
    self.outputs.clear();
    self.uniforms.clear();
    self.textures.clear();
    self.functions.clear();
}

pub fn defineFunction(self: *Self, arena: *std.heap.ArenaAllocator, function: coral.shaders.Function) coral.shaders.DefinitionError!void {
    const defined = try self.scope.define(arena, function);

    self.functions.append(defined) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineInput(self: *Self, arena: *std.heap.ArenaAllocator, input: coral.shaders.Input) coral.shaders.DefinitionError!void {
    const defined = try self.scope.define(arena, input);

    self.inputs.append(defined) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineOutput(self: *Self, arena: *std.heap.ArenaAllocator, output: coral.shaders.Output) coral.shaders.DefinitionError!void {
    const defined = try self.scope.define(arena, output);

    self.outputs.append(defined) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineTexture(self: *Self, arena: *std.heap.ArenaAllocator, input: coral.shaders.Texture) coral.shaders.DefinitionError!void {
    const defined = try self.scope.define(arena, input);

    self.textures.append(defined) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineUniform(self: *Self, arena: *std.heap.ArenaAllocator, input: coral.shaders.Uniform) coral.shaders.DefinitionError!void {
    const defined = try self.scope.define(arena, input);

    self.uniforms.append(defined) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn init(arena: *std.heap.ArenaAllocator) std.mem.Allocator.Error!Self {
    const scope = try arena.allocator().create(coral.shaders.Scope);

    scope.* = .{};

    return .{ .scope = scope };
}

pub fn parse(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!void {
    errdefer {
        self.clear();
    }

    while (source.skip(.newline) != .end) {
        const symbol = try self.scope.parse(arena, source);

        if (symbol.has(coral.shaders.Input)) |input| {
            self.inputs.append(input) catch |append_error| {
                return switch (append_error) {
                    error.Overflow => error.TooManySymbols,
                };
            };

            continue;
        }

        if (symbol.has(coral.shaders.Output)) |output| {
            self.outputs.append(output) catch |append_error| {
                return switch (append_error) {
                    error.Overflow => error.TooManySymbols,
                };
            };

            continue;
        }

        if (symbol.has(coral.shaders.Uniform)) |uniform| {
            self.uniforms.append(uniform) catch |append_error| {
                return switch (append_error) {
                    error.Overflow => error.TooManySymbols,
                };
            };

            continue;
        }

        if (symbol.has(coral.shaders.Texture)) |texture| {
            self.textures.append(texture) catch |append_error| {
                return switch (append_error) {
                    error.Overflow => error.TooManySymbols,
                };
            };

            continue;
        }

        if (symbol.has(coral.shaders.Function)) |function| {
            self.functions.append(function) catch |append_error| {
                return switch (append_error) {
                    error.Overflow => error.TooManySymbols,
                };
            };

            continue;
        }
    }
}

pub fn parseText(self: *Self, arena: *std.heap.ArenaAllocator, source_text: []const u8) std.mem.Allocator.Error!ParseResult {
    errdefer {
        self.clear();
    }

    var source = tokens.Stream.init(source_text);

    self.parse(arena, &source) catch |parse_error| {
        const arena_allocator = arena.allocator();

        return .{
            .failure = try switch (parse_error) {
                error.OutOfMemory => error.OutOfMemory,

                error.ImmutableStorage => coral.bytes.allocFormatted(arena_allocator, "{location}: attempt to modify an immutable value", .{
                    .location = source.location,
                }),

                error.TooManySymbols => coral.bytes.allocFormatted(arena_allocator, "{location}: number of definitions in the current scope exceeded", .{
                    .location = source.location,
                }),

                error.IncompatibleArguments => coral.bytes.allocFormatted(arena_allocator, "{location}: incompatible set of arguments", .{
                    .location = source.location,
                }),

                error.IncompatibleTypes => coral.bytes.allocFormatted(arena_allocator, "{location}: incompatible types", .{
                    .location = source.location,
                }),

                error.DuplicateIdentifier => coral.bytes.allocFormatted(arena_allocator, "{location}: {kind} {token} already defined", .{
                    .location = source.location,
                    .token = source.current,
                    .kind = std.meta.activeTag(source.current),
                }),

                error.UnexpectedToken => coral.bytes.allocFormatted(arena_allocator, "{location}: unexpected {token}", .{
                    .location = source.location,
                    .token = source.current,
                }),

                error.UndefinedIdentifier => coral.bytes.allocFormatted(arena_allocator, "{location}: undefined identifier {token}", .{
                    .location = source.location,
                    .token = source.current,
                }),
            },
        };
    };

    return .ok;
}
