const coral = @import("../coral.zig");

const std = @import("std");

pub const ExpectationError = error{
    UnexpectedToken,
};

pub const Kind = enum {
    end,
    unknown,
    newline,
    keyword_function,
    keyword_return,
    keyword_sample,
    keyword_float,
    keyword_float4x4,
    keyword_float2,
    keyword_float3,
    keyword_float4,
    keyword_int,
    keyword_let,
    keyword_var,
    keyword_pow,
    keyword_abs,
    keyword_sin,
    identifier,
    scalar,
    symbol_plus,
    symbol_minus,
    symbol_comma,
    symbol_arrow,
    symbol_equals,
    symbol_period,
    symbol_asterisk,
    symbol_paren_left,
    symbol_paren_right,
    symbol_brace_left,
    symbol_brace_right,
    symbol_double_equals,
    symbol_lesser_than,
    symbol_lesser_equals,
    symbol_greater_than,
    symbol_greater_equals,
    symbol_forward_slash,

    pub fn writeFormat(self: Kind, writer: coral.bytes.Writable) coral.bytes.ReadWriteError!void {
        try coral.bytes.writeAll(writer, self.name());
    }

    pub fn name(self: Kind) [:0]const u8 {
        return switch (self) {
            .end => "end",
            .unknown => "unknown",
            .newline => "newline",
            .keyword_function => "`function` keyword",
            .keyword_return => "`return` keyword",
            .keyword_sample => "`sample` keyword",
            .keyword_float => "`float` keyword",
            .keyword_float4x4 => "`float4x4` keyword",
            .keyword_float2 => "`float2` keyword",
            .keyword_float3 => "`float3` keyword",
            .keyword_float4 => "`float4` keyword",
            .keyword_int => "`int` keyword",
            .keyword_let => "`let` keyword",
            .keyword_var => "`var` keyword",
            .keyword_pow => "`pow` keyword",
            .keyword_abs => "`abs` keyword",
            .keyword_sin => "`sin` keyword",
            .identifier => "identifier",
            .scalar => "scalar literal",
            .symbol_plus => "`+`",
            .symbol_minus => "`-`",
            .symbol_comma => "`,`",
            .symbol_arrow => "`->`",
            .symbol_equals => "`=`",
            .symbol_period => "`.`",
            .symbol_asterisk => "`*`",
            .symbol_paren_left => "`(`",
            .symbol_paren_right => "`)`",
            .symbol_brace_left => "`{`",
            .symbol_brace_right => "`}`",
            .symbol_double_equals => "`==`",
            .symbol_lesser_than => "`<`",
            .symbol_lesser_equals => "`<=`",
            .symbol_greater_than => "`>`",
            .symbol_greater_equals => "`>=`",
            .symbol_forward_slash => "`/`",
        };
    }
};

pub const Location = struct {
    line_number: u32 = 1,
    column_number: u32 = 1,

    pub fn writeFormat(self: Location, writer: coral.bytes.Writable) coral.bytes.ReadWriteError!void {
        try coral.bytes.writeFormatted(writer, "({line}, {column})", .{
            .line = coral.utf8.cDec(self.line_number),
            .column = coral.utf8.cDec(self.column_number),
        });
    }
};

pub const Stream = struct {
    source: []const u8,
    current: Token,
    location: Location,
    depth: u32 = 0,

    pub fn isFinished(self: Stream) bool {
        return self.source.len == 0;
    }

    pub fn init(source: []const u8) Stream {
        return .{
            .source = source,
            .current = .end,
            .location = .{},
        };
    }

    pub fn next(self: *Stream) Token {
        const initial_len = self.source.len;

        self.current = consume(&self.source);

        switch (self.current) {
            .end => {
                self.depth = 0;
            },

            .symbol_brace_left => {
                self.depth += 1;
            },

            .symbol_brace_right => {
                self.depth -= 1;
            },

            .newline => {
                self.location.line_number += 1;
                self.location.column_number = 1;
            },

            else => {
                const consumed_len: u32 = @intCast(initial_len - self.source.len);

                self.location.column_number += consumed_len;
            },
        }

        return self.current;
    }

    pub fn skip(self: *Stream, skip_kind: Kind) Token {
        while (true) {
            if (std.meta.activeTag(self.next()) != skip_kind) {
                return self.current;
            }
        }
    }
};

pub const Token = union(Kind) {
    end,
    unknown: []const u8,
    newline,
    keyword_function,
    keyword_return,
    keyword_sample,
    keyword_float,
    keyword_float4x4,
    keyword_float2,
    keyword_float3,
    keyword_float4,
    keyword_int,
    keyword_let,
    keyword_var,
    keyword_pow,
    keyword_abs,
    keyword_sin,
    identifier: []const u8,
    scalar: []const u8,
    symbol_plus,
    symbol_minus,
    symbol_comma,
    symbol_arrow,
    symbol_equals,
    symbol_period,
    symbol_asterisk,
    symbol_paren_left,
    symbol_paren_right,
    symbol_brace_left,
    symbol_brace_right,
    symbol_double_equals,
    symbol_lesser_than,
    symbol_lesser_equals,
    symbol_greater_than,
    symbol_greater_equals,
    symbol_forward_slash,

    pub fn ExpectedAnyEnum(comptime expecteds: []const Kind) type {
        comptime var fields: [expecteds.len]std.builtin.Type.EnumField = undefined;

        inline for (0..expecteds.len) |i| {
            fields[i] = .{
                .name = @tagName(expecteds[i]),
                .value = i,
            };
        }

        return @Type(.{
            .@"enum" = .{
                .tag_type = @typeInfo(Kind).@"enum".tag_type,
                .fields = &fields,
                .decls = &.{},
                .is_exhaustive = true,
            },
        });
    }

    pub fn expect(self: Token, expected: Kind) ExpectationError!void {
        if (std.meta.activeTag(self) != expected) {
            return error.UnexpectedToken;
        }
    }

    pub fn expectAny(self: Token, comptime expecteds: []const Kind) ExpectationError!ExpectedAnyEnum(expecteds) {
        if (std.mem.indexOfScalar(Kind, expecteds, std.meta.activeTag(self))) |index| {
            return @enumFromInt(index);
        }

        return error.UnexpectedToken;
    }

    pub fn expectIdentifier(self: Token) ExpectationError![]const u8 {
        try self.expect(.identifier);

        return self.identifier;
    }

    pub fn expectScalar(self: Token) ExpectationError![]const u8 {
        try self.expect(.scalar);

        return self.scalar;
    }

    pub fn writeFormat(self: Token, writer: coral.bytes.Writable) coral.bytes.ReadWriteError!void {
        try switch (self) {
            .unknown => |unknown| coral.bytes.writeFormatted(writer, "`{name}`", .{
                .name = unknown,
            }),

            .identifier => |identifier| coral.bytes.writeFormatted(writer, "`{name}`", .{
                .name = identifier,
            }),

            .scalar => |literal| coral.bytes.writeFormatted(writer, "`{name}`", .{
                .name = literal,
            }),

            else => std.meta.activeTag(self).writeFormat(writer),
        };
    }
};

pub fn consume(source: *[]const u8) Token {
    var cursor = @as(usize, 0);

    defer {
        source.* = source.*[cursor..];
    }

    while (cursor < source.len) {
        switch (source.*[cursor]) {
            '#' => {
                cursor += 1;

                while (cursor < source.len and source.*[cursor] != '\n') {
                    cursor += 1;
                }
            },

            ' ', '\t' => {
                cursor += 1;
            },

            '\n' => {
                cursor += 1;

                return .newline;
            },

            '0'...'9' => {
                const begin = cursor;

                cursor += 1;

                while (cursor < source.len) {
                    switch (source.*[cursor]) {
                        '0'...'9' => {
                            cursor += 1;
                        },

                        '.' => {
                            cursor += 1;

                            while (cursor < source.len) {
                                switch (source.*[cursor]) {
                                    '0'...'9' => {
                                        cursor += 1;
                                    },

                                    else => {
                                        break;
                                    },
                                }
                            }

                            return .{ .scalar = source.*[begin..cursor] };
                        },

                        else => {
                            break;
                        },
                    }
                }

                return .{ .scalar = source.*[begin..cursor] };
            },

            'A'...'Z', 'a'...'z', '_' => {
                const begin = cursor;

                cursor += 1;

                while (cursor < source.len) {
                    switch (source.*[cursor]) {
                        '0'...'9', 'A'...'Z', 'a'...'z', '_' => cursor += 1,
                        else => break,
                    }
                }

                const identifier = source.*[begin..cursor];

                std.debug.assert(identifier.len != 0);

                const identifier_keyword_pairs = [_]struct { [:0]const u8, Kind }{
                    .{ "float", .keyword_float },
                    .{ "float2", .keyword_float2 },
                    .{ "float3", .keyword_float3 },
                    .{ "float4", .keyword_float4 },
                    .{ "float4x4", .keyword_float4x4 },
                    .{ "function", .keyword_function },
                    .{ "let", .keyword_let },
                    .{ "var", .keyword_var },
                    .{ "return", .keyword_return },
                    .{ "var", .keyword_var },
                    .{ "pow", .keyword_pow },
                    .{ "abs", .keyword_abs },
                    .{ "sin", .keyword_sin },
                    .{ "sample", .keyword_sample },
                };

                inline for (identifier_keyword_pairs) |pair| {
                    const keyword_identifier, const keyword = pair;

                    if (std.mem.eql(u8, identifier, keyword_identifier)) {
                        return keyword;
                    }
                }

                return .{ .identifier = identifier };
            },

            '{' => {
                cursor += 1;

                return .symbol_brace_left;
            },

            '}' => {
                cursor += 1;

                return .symbol_brace_right;
            },

            ',' => {
                cursor += 1;

                return .symbol_comma;
            },

            ')' => {
                cursor += 1;

                return .symbol_paren_right;
            },

            '(' => {
                cursor += 1;

                return .symbol_paren_left;
            },

            '/' => {
                cursor += 1;

                return .symbol_forward_slash;
            },

            '*' => {
                cursor += 1;

                return .symbol_asterisk;
            },

            '-' => {
                cursor += 1;

                if (cursor < source.len and source.*[cursor] == '>') {
                    cursor += 1;

                    return .symbol_arrow;
                }

                return .symbol_minus;
            },

            '+' => {
                cursor += 1;

                return .symbol_plus;
            },

            '=' => {
                cursor += 1;

                if (cursor < source.len and source.*[cursor] == '=') {
                    cursor += 1;

                    return .symbol_double_equals;
                }

                return .symbol_equals;
            },

            '<' => {
                cursor += 1;

                if (cursor < source.len and (source.*[cursor] == '=')) {
                    cursor += 1;

                    return .symbol_lesser_equals;
                }

                return .symbol_lesser_than;
            },

            '>' => {
                cursor += 1;

                if (cursor < source.len and (source.*[cursor] == '=')) {
                    cursor += 1;

                    return .symbol_greater_equals;
                }

                return .symbol_greater_than;
            },

            '.' => {
                cursor += 1;

                return .symbol_period;
            },

            else => {
                const begin = cursor;

                cursor += 1;

                while (cursor < source.len and !std.ascii.isWhitespace(source.*[cursor])) {
                    cursor += 1;
                }

                return .{ .unknown = source.*[begin..cursor] };
            },
        }
    }

    return .end;
}
