const coral = @import("./coral.zig");

const std = @import("std");

pub fn Dec(comptime Value: type) type {
    return struct {
        value: Value,
        format: *const DecFormat,

        const Self = @This();

        pub fn length(self: Self) usize {
            switch (@typeInfo(Value)) {
                .ComptimeInt => {
                    return self.length(@as(std.math.IntFittingRange(self.value, self.value), self.value));
                },

                .Int => |int| {
                    if (self.value == 0) {
                        return 1;
                    }

                    const digit_count = switch (int.signedness) {
                        .unsigned => @as(usize, @intCast(std.math.log10((self.value)))),
                        .signed => @as(usize, @intCast(std.math.log10((@abs(self.value))))),
                    };

                    const has_prefix = switch (int.signedness) {
                        .unsigned => switch (self.positive_prefix) {
                            .none => false,
                            .plus, .space => true,
                        },

                        .signed => self.value < 0,
                    };

                    const delimiter_space = ((digit_count - 1) / 3) * self.delimiter.len;

                    return digit_count + @intFromBool(has_prefix) + delimiter_space;
                },

                else => @compileError("Cannot retrieve the decimal string length of " ++ @typeName(Value)),
            }
        }

        pub fn writeFormat(self: Self, writer: coral.bytes.Writable) coral.bytes.ReadWriteError!void {
            if (self.value == 0) {
                return coral.bytes.writeAll(writer, switch (self.format.positive_prefix) {
                    .none => "0",
                    .plus => "+0",
                    .space => " 0",
                });
            }

            switch (@typeInfo(Value)) {
                .int => |int| {
                    const radix = 10;
                    var buffer = [_]u8{0} ** (1 + @max(int.bits, 1));
                    var buffer_start = buffer.len - 1;

                    {
                        var decomposable_value = self.value;

                        while (decomposable_value != 0) : (buffer_start -= 1) {
                            buffer[buffer_start] = @intCast(@mod(decomposable_value, radix) + '0');
                            decomposable_value = @divTrunc(decomposable_value, radix);
                        }
                    }

                    if (int.signedness == .unsigned and self.value < 0) {
                        buffer[buffer_start] = '-';
                    } else {
                        switch (self.format.positive_prefix) {
                            .none => buffer_start += 1,
                            .plus => buffer[buffer_start] = '+',
                            .space => buffer[buffer_start] = ' ',
                        }
                    }

                    return coral.bytes.writeAll(writer, buffer[buffer_start..]);
                },

                .float => |float| {
                    if (self.value < 0) {
                        try coral.bytes.writeAll(writer, "-");
                    }

                    const Int = std.meta.Int(.unsigned, float.bits);
                    const integer = @as(Int, @intFromFloat(self.value));

                    try self.writeFormat(writer, integer);
                    try coral.bytes.writeAll(writer, ".");
                    try self.writeFormat(writer, @as(Int, @intFromFloat((self.value - @as(Value, @floatFromInt(integer))) * 100)));
                },

                .comptime_int => {
                    const repacked = Dec(std.math.IntFittingRange(self.value, self.value)){
                        .value = self.value,
                        .format = self.format,
                    };

                    try repacked.writeFormat(writer);
                },

                else => @compileError("`" ++ @typeName(Value) ++ "` cannot be formatted to a decimal string"),
            }
        }
    };
}

pub const DecFormat = struct {
    delimiter: []const u8,
    positive_prefix: enum { none, plus, space },

    pub fn parse(self: DecFormat, utf8: []const u8, comptime Value: type) ?Value {
        if (utf8.len == 0) {
            return null;
        }

        switch (@typeInfo(Value)) {
            .Int => |int| {
                const has_sign = switch (utf8[0]) {
                    '-', '+', ' ' => true,
                    else => false,
                };

                var result = @as(Value, 0);

                for (@intFromBool(has_sign)..utf8.len) |index| {
                    const radix = 10;
                    const code = utf8[index];

                    switch (code) {
                        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' => {
                            const offset_code = coral.scalars.sub(code, '0') orelse {
                                return null;
                            };

                            result = coral.scalars.mul(result, radix) orelse {
                                return null;
                            };

                            result = coral.scalars.add(result, offset_code) orelse {
                                return null;
                            };
                        },

                        else => {
                            if (self.delimiter.len == 0 or !std.mem.eql(u8, self.delimiter, utf8[index..])) {
                                return null;
                            }
                        },
                    }
                }

                switch (int.signedness) {
                    .signed => {
                        return result * @as(Value, if (has_sign and utf8[0] == '-') -1 else 1);
                    },

                    .unsigned => {
                        if (has_sign and utf8[0] == '-') {
                            return null;
                        }

                        return result;
                    },
                }
            },

            .Float => {
                const has_sign = switch (utf8[0]) {
                    '-', '+', ' ' => true,
                    else => false,
                };

                // "-"
                if (has_sign and utf8.len == 1) {
                    return null;
                }

                const sign_offset = @intFromBool(has_sign);
                var has_decimal = utf8[sign_offset] == '.';

                // "-."
                if (has_decimal and (utf8.len == 2)) {
                    return null;
                }

                var result = @as(Value, 0);
                var factor = @as(Value, if (has_sign and utf8[0] == '-') -1 else 1);

                for (utf8[sign_offset + @intFromBool(has_decimal) .. utf8.len]) |code| {
                    switch (code) {
                        '.' => {
                            if (has_decimal) {
                                return null;
                            }

                            has_decimal = true;
                        },

                        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' => {
                            if (has_decimal) {
                                factor /= 10.0;
                            }

                            result = ((result * 10.0) + @as(Value, @floatFromInt(code - '0')));
                        },

                        else => return null,
                    }
                }

                return result * factor;
            },

            else => @compileError("`" ++ @typeName(Value) ++ "` cannot be parsed from a decimal string"),
        }
    }
};

test "decimal parsing" {
    const format = DecFormat{
        .delimiter = "",
        .positive_prefix = .none,
    };

    try std.testing.expectEqual(format.parse("69", i64), 69);
}

pub fn cDec(value: anytype) Dec(@TypeOf(value)) {
    return .{
        .value = value,

        .format = &.{
            .delimiter = "",
            .positive_prefix = .none,
        },
    };
}
