const coral = @import("coral");

const ona = @import("../ona.zig");

const std = @import("std");

dependants_edges: Map(BehaviorSet),
processed: Map(void),
state_readers: AccessMap,
state_writers: AccessMap,
blocking_work: BehaviorSet,
parallel_work: BehaviorSet,
parallel_work_ranges: coral.Stack(usize),

const AccessMap = coral.map.Hashed(*const coral.TypeId, BehaviorSet, coral.map.ptrTraits(*const coral.TypeId));

const BehaviorSet = coral.Stack(*const ona.App.Behavior);

fn Map(comptime Payload: type) type {
    return coral.map.Hashed(*const ona.App.Behavior, Payload, coral.map.ptrTraits(*const ona.App.Behavior));
}

const Self = @This();

pub const RunError = error{
    MissingDependency,
};

pub const TypeDependency = struct {
    id: *const coral.TypeId,
    is_read_only: bool,
};

pub fn deinit(self: *Self) void {
    self.processed.deinit();
    self.parallel_work.deinit();
    self.parallel_work_ranges.deinit();
    self.blocking_work.deinit();

    inline for (.{ &self.dependants_edges, &self.state_readers, &self.state_writers }) |map| {
        var entries = map.entries();

        while (entries.next()) |entry| {
            entry.value.deinit();
        }

        map.deinit();
    }
}

pub fn dependOnBehavior(self: *Self, app: *ona.App, dependant: *const ona.App.Behavior, dependency: *const ona.App.Behavior) std.mem.Allocator.Error!void {
    try self.insert(app, dependant);

    const edges = self.dependants_edges.get(dependant).?;

    if (std.mem.indexOfScalar(*const ona.App.Behavior, edges.values, dependency) == null) {
        try edges.pushGrow(dependency);
    }
}

pub fn dependOnType(self: *Self, app: *ona.App, dependant: *const ona.App.Behavior, dependency: TypeDependency) std.mem.Allocator.Error!void {
    var readers = self.state_readers.get(dependency.id) orelse (try self.state_readers.emplace(dependency.id, .empty)).?;

    if (std.mem.indexOfScalar(*const ona.App.Behavior, readers.values, dependant)) |index| {
        for (readers.values[0..index]) |reader| {
            try self.dependOnBehavior(app, dependant, reader);
        }

        for (readers.values[index + 1 ..]) |reader| {
            try self.dependOnBehavior(app, dependant, reader);
        }
    } else {
        for (readers.values) |reader| {
            try self.dependOnBehavior(app, dependant, reader);
        }

        try readers.pushGrow(dependant);
    }

    if (!dependency.is_read_only) {
        const writers = self.state_writers.get(dependency.id) orelse (try self.state_writers.emplace(dependency.id, .empty)).?;

        if (std.mem.indexOfScalar(*const ona.App.Behavior, writers.values, dependant)) |index| {
            for (writers.values[0..index]) |reader| {
                try self.dependOnBehavior(app, dependant, reader);
            }

            for (writers.values[index..]) |reader| {
                try self.dependOnBehavior(app, dependant, reader);
            }
        } else {
            for (writers.values) |reader| {
                try self.dependOnBehavior(app, dependant, reader);
            }

            try writers.pushGrow(dependant);
        }
    }
}

pub const empty = Self{
    .processed = .empty,
    .dependants_edges = .empty,
    .state_readers = .empty,
    .state_writers = .empty,
    .blocking_work = .empty,
    .parallel_work = .empty,
    .parallel_work_ranges = .empty,
};

pub fn insert(self: *Self, app: *ona.App, behavior: *const ona.App.Behavior) std.mem.Allocator.Error!void {
    self.processed.clear();

    if ((try self.dependants_edges.emplace(behavior, .empty)) != null) {
        try behavior.on_insertion(behavior, app, self);
    }
}

fn parallelRun(app: *ona.App, behaviors: []const *const ona.App.Behavior) void {
    for (behaviors) |behavior| {
        behavior.on_run(app);
    }
}

fn process(self: *Self, behavior: *const ona.App.Behavior, dependencies: BehaviorSet) !ona.Traits {
    var inherited_traits = behavior.traits;

    for (dependencies.values) |dependency| {
        const traits = try self.process(dependency, (self.dependants_edges.get(dependency) orelse {
            return error.MissingDependency;
        }).*);

        inherited_traits = inherited_traits.derived(traits);
    }

    if (try self.processed.emplace(behavior, {})) {
        try switch (inherited_traits.is_thread_unsafe) {
            true => self.blocking_work.pushGrow(behavior),
            false => self.parallel_work.pushGrow(behavior),
        };
    }

    return inherited_traits;
}

pub fn run(self: *Self, app: *ona.App, tasks: *coral.asio.TaskQueue) (std.mem.Allocator.Error || RunError)!void {
    if (self.processed.isEmpty()) {
        errdefer {
            self.processed.clear();
        }

        self.blocking_work.clear();
        self.parallel_work.clear();
        self.parallel_work_ranges.clear();

        var dependents_edges = self.dependants_edges.entries();

        while (dependents_edges.next()) |entry| {
            const parallel_work_offset = self.parallel_work.values.len;
            const flags = try self.process(entry.key, entry.value);

            if (flags.is_thread_unsafe) {
                std.debug.assert(parallel_work_offset == self.parallel_work.values.len);
            } else {
                const parallel_work_added = self.parallel_work.values.len - parallel_work_offset;

                if (parallel_work_added != 0) {
                    try self.parallel_work_ranges.pushGrow(parallel_work_added);
                }
            }
        }
    }

    var parallel_work_offset: usize = 0;

    for (self.parallel_work_ranges.values) |parallel_work_range| {
        const work = self.parallel_work.values[parallel_work_offset .. parallel_work_offset + parallel_work_range];

        try tasks.execute(coral.asio.CallTask(parallelRun).init(.{ app, work }));

        parallel_work_offset += parallel_work_range;
    }

    tasks.finish();

    for (self.blocking_work.values) |behavior| {
        behavior.on_run(app);
    }
}
