pub const App = @import("./App.zig");

pub const Assets = @import("./Assets.zig");

const coral = @import("coral");

pub const gfx = @import("./gfx.zig");

pub const hid = @import("./hid.zig");

const std = @import("std");

fn Channel(comptime Message: type) type {
    return struct {
        buffers: [2]coral.Stack(Message) = .{ .empty, .empty },
        swap_index: u1 = 0,
        ticks: u1 = 0,

        const Self = @This();

        pub fn bind(app: *App) std.mem.Allocator.Error!void {
            if (app.hasState(Self) == null) {
                try app.setState(Self{});
                try app.on(.post_update, .of(swap));
            }
        }

        pub fn deinit(self: *Self) void {
            for (&self.buffers) |*buffer| {
                buffer.deinit();
            }
        }

        fn swap(channel: Write(Self)) void {
            channel.ptr.ticks = coral.scalars.add(channel.ptr.ticks, 1) orelse 0;

            if (channel.ptr.ticks == 0) {
                channel.ptr.swap_index ^= 1;

                channel.ptr.buffers[channel.ptr.swap_index].clear();
            }
        }

        fn messages(self: Self) []const Message {
            return self.buffers[self.swap_index ^ 1].values;
        }

        fn push(self: *Self, message: Message) std.mem.Allocator.Error!void {
            try self.buffers[self.swap_index].pushGrow(message);
        }
    };
}

pub fn Read(comptime State: type) type {
    return struct {
        ptr: *const State,

        const Self = @This();

        pub fn init(state: *const State) Self {
            return .{
                .ptr = state,
            };
        }
    };
}

pub fn Receive(comptime Message: type) type {
    const TypedChannel = Channel(Message);

    return struct {
        channel: *const TypedChannel,

        const Self = @This();

        pub const bind = TypedChannel.bind;

        pub fn init(channel: *const TypedChannel) Self {
            return .{
                .channel = channel,
            };
        }

        pub fn messages(self: Self) []const Message {
            return self.channel.messages();
        }
    };
}

pub fn Send(comptime Message: type) type {
    const TypedChannel = Channel(Message);

    return struct {
        channel: *TypedChannel,

        const Self = @This();

        pub const bind = TypedChannel.bind;

        pub fn init(channel: *TypedChannel) Self {
            return .{
                .channel = channel,
            };
        }

        pub fn push(self: Self, message: Message) std.mem.Allocator.Error!void {
            try self.channel.push(message);
        }
    };
}

pub const Traits = packed struct {
    is_thread_unsafe: bool = false,

    pub fn derived(self: Traits, other: Traits) Traits {
        return .{
            .is_thread_unsafe = self.is_thread_unsafe or other.is_thread_unsafe,
        };
    }
};

pub fn Write(comptime State: type) type {
    return struct {
        ptr: *State,

        const Self = @This();

        pub fn init(state: *State) Self {
            return .{
                .ptr = state,
            };
        }
    };
}

pub const realtime_app = App.Setup.init(run_realtime_loop);

fn run_realtime_loop(app: *App) !void {
    // ext.SDL_SetLogPriorities(ext.SDL_LOG_PRIORITY_VERBOSE);
    // ext.SDL_SetLogOutputFunction(sdl_log, null);
    const updates_per_frame = 60.0;
    const target_frame_time = 1.0 / updates_per_frame;
    const time = app.hasState(App.Time).?;

    const virtual_thread_count = std.Thread.getCpuCount() catch 0;
    var tasks = try coral.asio.TaskQueue.init(virtual_thread_count / 2);

    defer {
        tasks.deinit();
    }

    try app.run(&tasks, .load);

    const ticks_initial = std.time.milliTimestamp();
    var ticks_previous = ticks_initial;
    var accumulated_time = @as(f64, 0);

    while (app.is_running) {
        const ticks_current = std.time.milliTimestamp();
        const milliseconds_per_second = 1000.0;

        accumulated_time += @as(f64, @floatFromInt(ticks_current - ticks_previous)) / milliseconds_per_second;
        time.elapsed = @as(f64, @floatFromInt(ticks_current - ticks_initial)) / milliseconds_per_second;
        ticks_previous = ticks_current;

        try app.run(&tasks, .pre_update);

        while (accumulated_time >= target_frame_time) : (accumulated_time -= target_frame_time) {
            try app.run(&tasks, .update);
        }

        try app.run(&tasks, .post_update);
        try app.run(&tasks, .render);
        try app.run(&tasks, .finish);
    }

    try app.run(&tasks, .exit);
}
